use std::collections::VecDeque;
use std::rc::Rc;

const MAXV: usize = 1000;

#[derive(Clone)]
struct Edgenode {
    #[allow(unused)]
    weight: i32, // Edge weight, if any

    y: usize,                     // Adjacency info
    next: Option<Rc<Edgenode>>, // next edge in list
}

pub struct Graph {
    edges: Vec<Option<Rc<Edgenode>>>, // adjacency info
    degree: Vec<i32>,                 // outdegree of each vertex
    nvertices: usize,                 // number of vertices in graph
    nedges: usize,                      // number of edges in graph
    directed: bool,                   // is the graph directed?
}

struct SearchState {
    // A vertex is `discovered` the first time we visit it.
    discovered: Vec<bool>,
    // A vertex is `processed` after we have traversed all outgoing
    // edges from it.
    processed: Vec<bool>,
    // Discovery relation. Needs to be signed cause we initialize it to -1
    parent: Vec<i32>,
    // Time tracking for DFS
    time: usize,
}

impl Graph {
    pub fn new(directed: bool) -> Self {
        Self {
            nvertices: 0,
            nedges: 0,
            directed,
            degree: vec![0; MAXV],
            edges: vec![None; MAXV],
        }
    }

    fn insert_edge(&mut self, x: usize, y: usize, directed: bool) {
        let p = Rc::new(Edgenode {
            y,
            weight: 0,
            next: self.edges[x].clone(),
        });
        self.edges[x] = Some(p);
        self.degree[x] += 1;
        if directed {
            self.nedges += 1;
        } else {
            self.insert_edge(y, x, true);
        }
    }

    pub fn print_graph(&self) {
        for i in 0..=self.nvertices {
            print!("{} -->", i);
            let mut p_opt = self.edges[i].clone();
            while let Some(p) = p_opt {
                print!(" {}", p.y);
                p_opt = p.next.clone();
            }
            println!();
        }
    }

    pub fn read_graph(&mut self, data: String) {
        self._read_graph(data, self.directed);
    }

    fn _read_graph(&mut self, data: String, directed: bool) {
        // reverse here so that we can pop() later
        let mut data_parsed: Vec<&str> = data.split(' ').rev().collect();

        // The first two digits in `data` are `nvertices` and the total number of edges
        // between them
        self.nvertices = data_parsed.pop().unwrap().parse().expect("invalid int");
        let num_edges: i32 = data_parsed.pop().unwrap().parse().expect("invalid int");

        // After that, parse edges as (x, y)
        for _ in 0..num_edges {
            let x: usize = data_parsed.pop().unwrap().parse().unwrap();
            let y: usize = data_parsed.pop().unwrap().parse().unwrap();
            self.insert_edge(x, y, directed);
        }
    }

    pub fn bfs(&self, start: usize) {
        let mut state = SearchState::new();
        let mut q: VecDeque<usize> = VecDeque::new();
        q.push_back(start);
        state.discovered[start] = true;

        while let Some(v) = q.pop_front() {
            state.processed[v] = true;
            let mut p_opt = &self.edges[v];
            while let Some(p) = p_opt {
                let y: usize = p.y;
                if !state.processed[y] || self.directed {
                    println!("process edge ({:?},{:?})", v, y);
                }
                if !state.discovered[y] {
                    q.push_back(y);
                    state.discovered[y] = true;
                    state.parent[y] = v.try_into().unwrap();
                }
                p_opt = &p.next;
            }
        }
    }

    pub fn dfs(&self, start: usize) {
        let mut state = SearchState::new();
        let mut q: Vec<usize> = Vec::new();
        q.push(start);
        state.discovered[start] = true;

        while let Some(v) = q.pop() {
            let mut p_opt = &self.edges[v];
            while let Some(p) = p_opt {
                let y: usize = p.y;
                if !state.discovered[y] {
                    state.discovered[y] = true;
                    state.parent[y] = v.try_into().unwrap();
                    q.push(y);
                    println!("process edge ({:?},{:?})", v, y);
                } else if !state.processed[y] || self.directed {
                    println!("process edge ({:?},{:?})", v, y);
                }
                if state.parent[y] != v.try_into().unwrap() {
                    println!("cycle from {:?} to {:?}", v, y);
                }
                p_opt = &p.next;
            }
            state.processed[v] = true;
        }
    }
}

impl SearchState {
    pub fn new() -> Self {
        SearchState {
            discovered: vec![false; MAXV],
            processed: vec![false; MAXV],
            parent: vec![-1; MAXV],
            time: 0,
        }
    }
}
